import numpy as np
from pathlib import Path
import os
import glob
from natsort import os_sorted
import scipy.io as spio
import h5py
import matplotlib.pyplot as plt


def lister_it(in_list, keep_strings=None, remove_string=None):
    """

    Parameters
    ----------
    in_list : list
    keep_strings : list
    remove_string : list

    Returns
    -------

    """
    if isinstance(keep_strings, str):
        keep_strings = [keep_strings]
    if isinstance(remove_string, str):
        remove_string = [remove_string]

    if keep_strings is None:
        new_list = in_list
    else:
        new_list = []
        for L in in_list:
            for k in keep_strings:
                if k in L:
                    new_list.append(L)

    if remove_string is None:
        new_list_2 = new_list
    else:
        new_list_2 = []
        for L in new_list:
            for k in remove_string:
                if k not in L:
                    new_list_2.append(L)
    return new_list_2


def plot_pole_tracking_max_vals(h5_file):
    with h5py.File(h5_file, 'r') as hf:
        for i, k in enumerate(hf['max_val_stack'][:]):
            plt.plot()


def get_class_info(c, include_underscores=False):
    """look at the variables and methods of a class, prints an aligned list with their respective 'type'

    Parameters
    ----------
    c : class
        calss variable
    include_underscores : bool
        display the secret methods/variables starting with '_' (Default value = False)

    Returns
    -------
    type
        prints out a nice list of the modethods

    """
    names = []
    type_to_print = []
    for k in dir(c):
        if include_underscores is False and k[0] != '_':
            tmp1 = str(type(eval('c.' + k)))
            type_to_print.append(tmp1.split("""'""")[-2])
            names.append(k)
        elif include_underscores:
            tmp1 = str(type(eval('c.' + k)))
            type_to_print.append(tmp1.split("""'""")[-2])
            names.append(k)
    len_space = ' ' * max(len(k) for k in names)
    for k1, k2 in zip(names, type_to_print):
        k1 = (k1 + len_space)[:len(len_space)]
        print(k1 + ' type->   ' + k2)


def group_consecutives(vals, step=1):
    """

    Parameters
    ----------
    vals :
        
    step :
         (Default value = 1)

    Returns
    -------

    """
    run = []
    run_ind = []
    result = run
    result_ind = run_ind
    expect = None
    for k, v in enumerate(vals):
        if v == expect:
            if not (np.isnan(v)):
                # print(v)
                # print(expect)
                run.append(v)
                run_ind.append(k)
        else:
            if not (np.isnan(v)):
                run = [v]
                run_ind = [k]
                result.append(run)
                result_ind.append(run_ind)
        expect = v + step
    # print(result)
    if result == []:
        pass
    elif result[0] == []:
        result = result[1:]
        result_ind = result_ind[1:]
    return result, result_ind


def get_h5s(base_dir):
    """

    Parameters
    ----------
    base_dir :
        

    Returns
    -------

    """
    H5_file_list = []
    for path in Path(base_dir + '/').rglob('*.h5'):
        H5_file_list.append(str(path.parent) + os.path.sep + path.name)
    H5_file_list.sort()
    print_list_with_inds(H5_file_list)
    return H5_file_list


def check_if_file_lists_match(H5_list_LAB, H5_list_IMG):
    """

    Parameters
    ----------
    H5_list_LAB :
        
    H5_list_IMG :
        

    Returns
    -------

    """
    for h5_LAB, h5_IMG in zip(H5_list_LAB, H5_list_IMG):
        try:
            assert h5_IMG.split(os.path.sep)[-1] in h5_LAB
        except:
            print('DO NOT CONTINUE --- some files do not match on your lists try again')
            assert (1 == 0)
    print('yay they all match!')


def print_list_with_inds(list_in):
    """

    Parameters
    ----------
    list_in :
        

    Returns
    -------

    """
    _ = [print(str(i) + ' ' + k.split(os.path.sep)[-1]) for i, k in enumerate(list_in)]


def get_model_list(model_save_dir):
    """

    Parameters
    ----------
    model_save_dir :
        

    Returns
    -------

    """
    print('These are all the models to choose from...')
    model_2_load_all = glob.glob(model_save_dir + '/*.ckpt')
    print_list_with_inds(model_2_load_all)
    return model_2_load_all


def recursive_dir_finder(base_path, search_term):
    """enter base directory and search term to find all the directories in base directory
      with files matching the search_term. output a sorted list of directories.
      e.g. -> recursive_dir_finder('/content/mydropbox/', '*.mp4')

    Parameters
    ----------
    base_path :
        
    search_term :
        

    Returns
    -------

    """
    matching_folders = []
    for root, dirs, files in os.walk(base_path):
        if glob.glob(root + '/' + search_term):
            matching_folders.append(root)
    try:
        matching_folders = os_sorted(matching_folders)
    except:
        matching_folders = sorted(matching_folders)
    return matching_folders


def get_model_list(model_save_dir):
    """

    Parameters
    ----------
    model_save_dir :
        

    Returns
    -------

    """
    print('These are all the models to choose from...')
    model_2_load_all = sorted(glob.glob(model_save_dir + '/*.ckpt'))
    # print(*model_2_load_all, sep = '\n')
    _ = [print(str(i) + ' ' + k.split(os.path.sep)[-1]) for i, k in enumerate(model_2_load_all)]
    return model_2_load_all


def get_files(base_dir, search_term):
    """

    Parameters
    ----------
    base_dir :
        
    search_term :
        

    Returns
    -------

    """
    file_list = []
    for path in Path(base_dir + '/').rglob(search_term):
        file_list.append(str(path.parent) + os.path.sep + path.name)
    file_list.sort()
    return file_list


'''
these below 3 function used to load mat files into dict easily was found and copied directly from 
https://stackoverflow.com/questions/7008608/scipy-io-loadmat-nested-structures-i-e-dictionaries
and contributed by user -mergen 

to the best of my knowledge code found on stackoverflow is under the creative commons license and as such is legal to 
use in my package. contact phillip.maire@gmail.com if you have any questions. 
'''


def loadmat(filename):
    """this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects

    Parameters
    ----------
    filename :
        

    Returns
    -------

    """
    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


def _check_keys(dict):
    """checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries

    Parameters
    ----------
    dict :
        

    Returns
    -------

    """
    for key in dict:
        if isinstance(dict[key], spio.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict


def _todict(matobj):
    """A recursive function which constructs from matobjects nested dictionaries

    Parameters
    ----------
    matobj :
        

    Returns
    -------

    """
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, spio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict



import h5py
from whacc import utils

import matplotlib.pyplot as plt
import numpy as np

H5_list = utils.get_h5s('/Users/phil/Dropbox/Autocurator/testing_data/MP4s/')

H5_FILE = H5_list[0]
with h5py.File(H5_FILE, 'r') as hf:
    for k in hf.keys():
        print(k)
    print(hf['trial_nums_and_frame_nums'][:])


# with h5py.File(H5_FILE, 'r') as hf:
#     plt.plot(hf['in_range'][:])
#     print(hf['in_range'][:])


with h5py.File(H5_FILE, 'r') as hf:
    # for k in hf['trial_nums_and_frame_nums'][1, :]:
    cumsum_frames = np.concatenate((np.asarray([0]), np.cumsum(hf['trial_nums_and_frame_nums'][1, :])))
    tot_frames = np.sum(hf['trial_nums_and_frame_nums'][1, :])
start_pole = 2000
stop_pole = 3000

b = np.vstack((start_pole+cumsum_frames[:-1], cumsum_frames[1:]-1)).astype('int')
b = np.min(b, axis = 0)
a = np.vstack((stop_pole+cumsum_frames[:-1], cumsum_frames[1:])).astype('int')
a = np.min(a, axis = 0)

keep_mask = np.zeros(tot_frames.astype('int'))
for k1, k2 in zip(b, a):
    keep_mask[k1:k2] = 1
plt.plot(keep_mask)



H5_file_name = '/Users/phil/Dropbox/Autocurator/data/samsons_subsets/train_and_validate/AH1131X25032020ses335_subset.h5'
label_read_key = 'labels'
label_write_key = 'labels'
from whacc.touch_curation_GUI import touch_gui
touch_gui(H5_file_name, label_read_key, label_write_key)





from whacc import utils
from whacc import image_tools
from keras.preprocessing.image import ImageDataGenerator
import h5py
import matplotlib.pyplot as plt
from tqdm import tqdm

from imgaug import augmenters as iaa  # optional program to further augment data

h5_subset_file = '/Users/phil/Dropbox/Autocurator/data/samsons_subsets/train_and_validate/training_set.h5'

datagen = ImageDataGenerator(rotation_range=360,  #
                             width_shift_range=.05,  #
                             height_shift_range=.05,  #
                             shear_range=.05,  #
                             zoom_range=.1,
                             brightness_range=[0.2, 1.1])  #

with h5py.File(h5_subset_file, 'r') as hf:
    test_img = hf['images'][0]  # grab a single image

aug_ims, _ = image_tools.augment_helper(datagen, 30, 0 ,test_img, -1)  # make 99 of the augmented images and output 1 of th original
# '-1' is the label associate ith the image, we dont care becasue we just want the iamge so it doenst matter what it is here

plt.figure(figsize=[5, 5])
plt.imshow(image_tools.img_unstacker(aug_ims, 10), )
plt.show()

gaussian_noise = iaa.AdditiveGaussianNoise(5, 10)  # range of 5 to 10
aug_ims_2 = gaussian_noise.augment_images(aug_ims)

plt.figure(figsize=[5, 5])
plt.imshow(image_tools.img_unstacker(aug_ims_2, 10))
plt.show()

# once we are happy with out augmentation process we can make an augmented H5 file using class
new_H5_file = h5_subset_file.split('.')[0] + '_AUG.h5' # create new file name based on teh original H5 name

h5creator = image_tools.h5_iterative_creator(new_H5_file, overwrite_if_file_exists=True,
                                             close_and_open_on_each_iteration=True)

utils.get_class_info(h5creator)
with h5py.File(h5_subset_file, 'r') as hf:
    for image, label in zip(tqdm(hf['images'][:]), hf['labels'][:]):
        aug_img_stack, labels_stack = image_tools.augment_helper(datagen, 30, 0, image, label)
        # aug_img_stack = gaussian_noise.augment_images(aug_img_stack) # optional
        h5creator.add_to_h5(aug_img_stack, labels_stack)






import h5py
import numpy as numpy
import matplotlib.pyplot as plt
from whacc.utils import *


class subset_h5_generator:
    """ """
    def __init__(self, h5_img_file, label_key):
        self.h5_img_file = h5_img_file
        self.label_key = label_key
        with h5py.File(self.h5_img_file, 'r') as F:
            self.labels = F[label_key][:]

    def save_subset_h5_file(self, file_save_name = None):
        """

        Parameters
        ----------
        file_save_name :
             (Default value = None)

        Returns
        -------

        """
        if file_save_name is None:
            file_save_name = self.h5_img_file.split('.h5')[0] + '_subset.h5'
        all_labels = self.labels
        all_inds = self.all_inds
        try:
            os.remove(file_save_name)
            print('Deleting file with same name')
        except:
            pass
        with h5py.File(self.h5_img_file, 'r') as F:
            images = numpy.asarray([F['images'][k] for k in all_inds])
            labels = numpy.asarray([all_labels[k] for k in all_inds.astype('int')])
            in_range = numpy.asarray([F['in_range'][k] for k in all_inds])
            with h5py.File(file_save_name, 'w') as hf:  # auto close in case of failure using 'with'
                hf.create_dataset('images', data=images)
                hf.create_dataset('labels', data=labels)
                hf.create_dataset('in_range', data=in_range)
                hf.create_dataset('all_inds', data=all_inds)
                hf.create_dataset('retrain_H5_info', data=[str(self.retrain_H5_info).encode("ascii", "ignore")])
                hf.close()
                print('finished with ' + file_save_name)

    def plot_pole_grab(self, im_stack, fig_size=None):
        """

        Parameters
        ----------
        im_stack :
            
        fig_size :
             (Default value = None)

        Returns
        -------

        """
        if fig_size is None:
            width_plt = self.retrain_H5_info['num_high_prob_past_max_y'] + self.retrain_H5_info['seg_len_look_dist'] * 2
            plt.figure(figsize=[width_plt, 5])
        else:
            plt.figure(figsize=fig_size)
        _ = plt.imshow(im_stack)

    def keep_only_pole_up_times(self, start_pole, stop_pole):
        """

        Parameters
        ----------
        start_pole :
            
        stop_pole :
            

        Returns
        -------

        """
        with h5py.File(self.h5_img_file, 'r') as hf:
            cumsum_frames = np.concatenate((np.asarray([0]), np.cumsum(hf['trial_nums_and_frame_nums'][1, :])))
            tot_frames = np.sum(hf['trial_nums_and_frame_nums'][1, :])

        b = np.vstack((start_pole + cumsum_frames[:-1], cumsum_frames[1:] - 1)).astype('int')
        b = np.min(b, axis=0)
        a = np.vstack((stop_pole + cumsum_frames[:-1], cumsum_frames[1:])).astype('int')
        a = np.min(a, axis=0)

        keep_mask = np.zeros(tot_frames.astype('int'))
        for k1, k2 in zip(b, a):
            keep_mask[k1:k2] = 1
        return keep_mask

    def get_example_segments(self,
                             seg_len_before_touch=10,
                             seg_len_after_touch=10,
                             min_y=.2,
                             max_y=.8,
                             num_to_sample=40,
                             min_seg_size=6,
                             start_and_stop_pole_times=None):
        """

        Parameters
        ----------
        seg_len_before_touch :
             (Default value = 10)
        seg_len_after_touch :
             (Default value = 10)
        min_y :
             (Default value = .2)
        max_y :
             (Default value = .8)
        num_to_sample :
             (Default value = 40)
        min_seg_size :
             (Default value = 6)
        start_and_stop_pole_times :
             (Default value = None)

        Returns
        -------

        """
        labels = self.labels
        if start_and_stop_pole_times is not None:
            pole_mask = self.keep_only_pole_up_times(start_pole=start_and_stop_pole_times[0],
                                                     stop_pole=start_and_stop_pole_times[1])
            labels = labels * pole_mask
        segs = numpy.where(labels > min_y)[0]
        chunks_tmp, chunk_inds = group_consecutives(segs, step=1)
        # print(len(chunks))
        chunks = []
        for i, k in enumerate(chunks_tmp):
            if len(k) >= min_seg_size:
                chunks.append(chunks_tmp[i])
                # print('popped it')
        # print(len(chunks))
        self.chunks = chunks  ###remove
        good_up_segs = []
        good_down_segs = []
        for k in chunks:
            try:
                up = labels[k[0] - seg_len_before_touch:k[0]]
                down = labels[k[-1] - 1:k[-1] + seg_len_before_touch - 1]
                assert (len(up) == seg_len_before_touch)
                assert (len(down) == seg_len_before_touch)
                good_up_segs.append(numpy.min(up) <= min_y)
                good_down_segs.append(numpy.min(down) <= min_y)
            except:  # if we are on the edges just toss these
                good_up_segs.append(False)
                good_down_segs.append(False)
        good_up_segs = numpy.where(good_up_segs)[0]
        good_down_segs = numpy.where(good_down_segs)[0]
        a = numpy.random.choice(good_up_segs, size=num_to_sample, replace=False)
        up_start = [chunks[k][0] + 1 for k in a]

        a = numpy.random.choice(good_down_segs, size=num_to_sample, replace=False)
        down_start = [chunks[k][-1] + 1 for k in a]

        all_inds = numpy.asarray([])
        onset_list = []
        offset_list = []
        for k1, k2 in zip(up_start, down_start):
            onset_list.append(numpy.asarray(range(k1 - seg_len_before_touch, k1 + seg_len_after_touch)))
            all_inds = numpy.concatenate((all_inds, onset_list[-1]))
            offset_list.append(numpy.asarray(range(k2 - 1 - seg_len_after_touch, k2 - 1 + seg_len_before_touch)))
            all_inds = numpy.concatenate((all_inds, offset_list[-1]))
        retrain_H5_info = {'seg_len_look_dist': seg_len_before_touch,
                           'min_y': min_y,
                           'max_y': max_y,
                           'num_to_sample': num_to_sample,
                           'num_high_prob_past_max_y': seg_len_after_touch}
        inds_2_add = numpy.linspace(0, 50 - 1, ).astype(int)
        all_inds = numpy.concatenate((numpy.float64(inds_2_add), all_inds))
        self.all_inds = all_inds
        self.onset_list = onset_list
        self.offset_list = offset_list
        self.retrain_H5_info = retrain_H5_info
        # return all_inds, onset_list, offset_list, retrain_H5_info

    def get_img_stack(self, frame_inds, h5_img_key='images'):
        """

        Parameters
        ----------
        frame_inds :
            
        h5_img_key :
             (Default value = 'images')

        Returns
        -------

        """

        im_stack = None
        for k in frame_inds:
            with h5py.File(self.h5_img_file, 'r') as h:
                tmp_img = numpy.asarray(h[h5_img_key][k])
                if im_stack is None:
                    im_stack = tmp_img
                else:
                    im_stack = numpy.hstack((im_stack, tmp_img))
        return im_stack

    def plot_all_onset_or_offset(self, up_or_down_list, fig_size=None, h5_img_file=None):
        """

        Parameters
        ----------
        up_or_down_list :
            
        fig_size :
             (Default value = None)
        h5_img_file :
             (Default value = None)

        Returns
        -------

        """
        if h5_img_file is None:
            h5_img_file = self.h5_img_file

        for i, k in enumerate(up_or_down_list):
            if i == 0:
                im_stack = self.get_img_stack(k)
            else:
                im_stack = numpy.vstack((im_stack, self.get_img_stack(k)))
        self.plot_pole_grab(im_stack, fig_size)



import h5py
import matplotlib.pyplot as plt
from whacc import image_tools


h5_file = "/Users/phil/Dropbox/Autocurator/data/samsons_subsets/test/AH1159X27012021xS414_subset_2.h5"
with h5py.File(h5_file, 'r') as h:
    plt.imshow(image_tools.img_unstacker(h['images'][:], 21))


